/*
 * mCaptcha - A proof of work based DoS protection system
 * Copyright © 2023 Aravinth Manivannan <realravinth@batsense.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
use std::collections::BTreeMap;
use std::collections::BTreeSet;
use std::collections::HashMap;

use actix_web::get;
use actix_web::post;
use actix_web::web;
use actix_web::web::Data;
use actix_web::Responder;
use openraft::error::Infallible;
use openraft::BasicNode;
use openraft::RaftMetrics;
use web::Json;

use crate::app::DcacheApp;
use crate::DcacheNodeId;

#[post("/add-learner")]
pub async fn add_learner(
    app: Data<DcacheApp>,
    req: Json<(DcacheNodeId, String)>,
) -> actix_web::Result<impl Responder> {
    let node_id = req.0 .0;
    let node = BasicNode {
        addr: req.0 .1.clone(),
    };
    let res = app.raft.add_learner(node_id, node, true).await;

    Ok(Json(res))
}

#[post("/change-membership")]
pub async fn change_membership(
    app: Data<DcacheApp>,
    req: Json<BTreeSet<DcacheNodeId>>,
) -> actix_web::Result<impl Responder> {
    let res = app.raft.change_membership(req.0, false).await;
    Ok(Json(res))
}

#[post("/init")]
pub async fn init(app: Data<DcacheApp>) -> actix_web::Result<impl Responder> {
    let mut nodes = BTreeMap::new();
    nodes.insert(
        app.id,
        BasicNode {
            addr: app.addr.clone(),
        },
    );
    let res = app.raft.initialize(nodes).await;
    Ok(Json(res))
}

#[get("/metrics")]
pub async fn metrics(app: Data<DcacheApp>) -> actix_web::Result<impl Responder> {
    let metrics = app.raft.metrics().borrow().clone();
    let res: Result<RaftMetrics<DcacheNodeId, BasicNode>, Infallible> = Ok(metrics);
    Ok(Json(res))
}

use tokio::sync::mpsc;

#[derive(Debug)]
pub enum HealthStatus {
    Healthy(DcacheNodeId),
    Down(DcacheNodeId),
}

pub struct HealthMetrics;

impl HealthMetrics {
    pub async fn spawn(
        app: Data<DcacheApp>,
        threshold: usize,
        mut rx: mpsc::Receiver<HealthStatus>,
    ) -> tokio::task::JoinHandle<()> {
        let mut current_error_rate = HashMap::new();
        let fut = async move {
            while let Some(msg) = rx.recv().await {
                match msg {
                    HealthStatus::Healthy(id) => {
                        current_error_rate.insert(id, 0);
                    }
                    HealthStatus::Down(id) => {
                        if let Some(existing_count) =
                            current_error_rate.get(&id).map(|s| s.to_owned())
                        {
                            current_error_rate.insert(id, existing_count + 1);
                            if existing_count == threshold {
                                let cluster_metrics = app.raft.metrics().borrow().clone();
                                let mut new_nodes: Vec<DcacheNodeId> = Vec::new();
                                for node in cluster_metrics.membership_config.nodes() {
                                    if *node.0 == id {
                                        continue;
                                    }

                                    new_nodes.push(*node.0);
                                }

                                let res =
                                    app.raft.change_membership(new_nodes, false).await.unwrap();
                            }
                        } else {
                            current_error_rate.insert(id, 1);
                        }
                    }
                }
            }
        };

        tokio::spawn(fut)
    }
}

//#[get("/self/remove/{id}")]
//pub async fn remove_node(app: Data<DcacheApp>, id: web::Path<u64>) -> actix_web::Result<impl Responder> {
//    let cluster_metrics = app.raft.metrics().borrow().clone();
//    let remote_id: u64 = 3;
//    let mut new_nodes: Vec<DcacheNodeId> = Vec::new();
//    for node in cluster_metrics.membership_config.nodes() {
//        if *node.0 == remote_id {
//            continue;
//        }
//
//        new_nodes.push(*node.0);
//    }
//
//    let res = app.raft.change_membership(new_nodes, false).await;
//    Ok(Json(res))
//}
